#include "blBoundaryLayer.h"
#include "blClosures.h"
#include <cmath>
#include <iomanip>
#include <iostream>

using namespace blast;

/**
 * @brief Construct a new BoundaryLayer::BoundaryLayer object.
 *
 * @param _xtrF Forced transition location (default=-1; free transition).
 */
BoundaryLayer::BoundaryLayer(double _xtrF, std::string _name)
{
    xtrF = _xtrF;
    name = _name;
    xtr = 1.0;
    closSolver = new Closures();
    transitionNode = 0;
}

BoundaryLayer::~BoundaryLayer()
{
    delete closSolver;
    std::cout << "~BoundaryLayer()\n";
}

void BoundaryLayer::setMesh(const std::vector<double> &_x, const std::vector<double> &_y,
                            const std::vector<double> &_z, const double &_chord,
                            const double &_xMin, const std::vector<int> &_rows,
                            const std::vector<int> &_no)
{
    if (_x.size() != _y.size() || _x.size() != _z.size())
        throw std::runtime_error("Mesh coordinates are not consistent.");
    nNodes = _x.size();
    nElms = nNodes - 1;
    if (_rows.size() == 0)
    {
        rows.resize(0);
        for (size_t i = 0; i < nNodes; ++i)
            rows.push_back(static_cast<int>(i));
    }
    else if (_rows.size() != nNodes)
        throw std::runtime_error("Row vector is not consistent.");
    else
    {
        rows = _rows;
    }
    // Elements
    if (_no.size() == 0)
    {
        no.resize(0);
        for (size_t i = 0; i < nElms; ++i)
            no.push_back(static_cast<int>(i));
    }
    else if (_no.size() != nElms)
        throw std::runtime_error("No vector is not consistent.");
    else
    {
        no = _no;
    }

    x = _x;
    y = _y;
    z = _z;
    chord = _chord;
    xMin = _xMin;

    // Compute the local coordinate.
    this->computeLocalCoordinates();
    this->reset();
}

void BoundaryLayer::computeLocalCoordinates()
{
    loc.resize(nNodes, 0.);
    alpha.resize(nElms, 0.);
    dx.resize(nElms, 0.);

    for (size_t iPoint = 0; iPoint < nElms; ++iPoint)
    {
        alpha[iPoint] =
            std::atan2(y[iPoint + 1] - y[iPoint], x[iPoint + 1] - x[iPoint]);
        // dx = sqrt(delta x ^2 + delta y ^2).
        dx[iPoint] =
            std::sqrt((x[iPoint + 1] - x[iPoint]) * (x[iPoint + 1] - x[iPoint]) +
                      (y[iPoint + 1] - y[iPoint]) * (y[iPoint + 1] - y[iPoint]));
        loc[iPoint + 1] = loc[iPoint] + dx[iPoint];
    }

    // Compute the x/c coordinate.
    if (chord <= 0.)
        throw std::runtime_error("Chord length is zero or negative.");
    xoc.resize(nNodes, 0.);
    for (size_t iPoint = 0; iPoint < nNodes; ++iPoint)
        xoc[iPoint] = (x[iPoint] - xMin) / chord;
}

void BoundaryLayer::reset()
{
    u.resize(nNodes * getnVar(), 0.);
    Ret.resize(nNodes, 0.);
    cd.resize(nNodes, 0.);
    cf.resize(nNodes, 0.);
    Hstar.resize(nNodes, 0.);
    Hstar2.resize(nNodes, 0.);
    Hk.resize(nNodes, 0.);
    ctEq.resize(nNodes, 0.);
    us.resize(nNodes, 0.);
    deltaStar.resize(nNodes, 0.);
    delta.resize(nNodes, 0.);
    deltaStar.resize(nNodes, 0.);
    regime.resize(nNodes, 0);
    blowingVelocity.resize(nElms, 0.);

    Me.resize(nNodes, 0.);
    vt.resize(nNodes, 0.);
    rhoe.resize(nNodes, 0.);
    deltaStarExt.resize(nNodes, 0.);
    vtExt.resize(nNodes, 0.);
    xxExt.resize(nNodes, 0.);

    transitionNode = 0;
    xtr = 1.0;
}

/**
 * @brief Set the boundary condition at the stagnation point.
 *
 * @param Re Freestream Reynolds number.
 */
void BoundaryLayer::setStagnationBC(double Re)
{
    double dv0 = std::abs((vt[1] - vt[0]) / (loc[1] - loc[0]));
    if (dv0 > 0)
        u[0] = sqrt(0.075 / (Re * dv0)); // Theta
    else
        u[0] = 1e-6; // Theta
    u[1] = 2.23;     // H
    u[2] = 0.;       // N
    u[3] = vt[0];    // ue
    u[4] = 0.;       // Ctau

    Ret[0] = u[3] * u[0] * Re;
    if (Ret[0] < 1.)
    {
        Ret[0] = 1.;
        u[3] = Ret[0] / (u[0] * Re);
    }
    deltaStar[0] = u[0] * u[1];

    std::vector<double> lamParam(8, 0.);
    closSolver->laminarClosures(lamParam, u[0], u[1], Ret[0], Me[0], name);

    Hstar[0] = lamParam[0];
    Hstar2[0] = lamParam[1];
    Hk[0] = lamParam[2];
    cf[0] = lamParam[3];
    cd[0] = lamParam[4];
    delta[0] = lamParam[5];
    ctEq[0] = lamParam[6];
    us[0] = lamParam[7];
}

/**
 * @brief Set the boundary condition at the first wake point.
 *
 * @param Re Freestream Reynolds number
 * @param UpperCond Variables at the last point on the upper side.
 * @param LowerCond Variables at the last point on the lower side.
 */
void BoundaryLayer::setWakeBC(double Re, std::vector<double> &UpperCond,
                              std::vector<double> &LowerCond)
{
    if (name != "wake")
        std::cout << "Warning: Imposing wake boundary condition on " << name
                  << std::endl;
    u[0] = UpperCond[0] + LowerCond[0]; // (Theta_up+Theta_low).
    u[1] = (UpperCond[0] * UpperCond[1] + LowerCond[0] * LowerCond[1]) /
           u[0];  // ((Theta*H)_up+(Theta*H)_low)/(Theta_up+Theta_low).
    u[2] = 9.;    // Amplification factor.
    u[3] = vt[0]; // Inviscid velocity.
    u[4] = (UpperCond[0] * UpperCond[4] + LowerCond[0] * LowerCond[4]) /
           u[0]; // ((Ct*Theta)_up+(Theta)_low)/(Ct*Theta_up+Theta_low).

    Ret[0] = u[3] * u[0] * Re; // Reynolds number based on the momentum thickness.

    if (Ret[0] < 1.)
    {
        Ret[0] = 1.;
        u[3] = Ret[0] / (u[0] * Re);
    }
    deltaStar[0] = u[0] * u[1];

    // Laminar closures.
    std::vector<double> lamParam(8, 0.);
    closSolver->laminarClosures(lamParam, u[0], u[1], Ret[0], Me[0], name);
    Hstar[0] = lamParam[0];
    Hstar2[0] = lamParam[1];
    Hk[0] = lamParam[2];
    cf[0] = lamParam[3];
    cd[0] = lamParam[4];
    delta[0] = lamParam[5];
    ctEq[0] = lamParam[6];
    us[0] = lamParam[7];

    for (size_t k = 0; k < nNodes; ++k)
        regime[k] = 1;
}

/**
 * @brief Compute the blowing velocity in each station of the region.
 *
 */
void BoundaryLayer::computeBlowingVelocity()
{
    blowingVelocity.resize(nNodes - 1, 0.);
    for (size_t i = 1; i < nNodes; ++i)
    {
        blowingVelocity[i - 1] = (rhoe[i] * vt[i] * deltaStar[i] -
                                  rhoe[i - 1] * vt[i - 1] * deltaStar[i - 1]) /
                                 (rhoe[i] * (loc[i] - loc[i - 1]));
        if (std::isnan(blowingVelocity[i - 1]))
        {
            if (rhoe[i] == 0.)
                std::cout << "Density is zero at point " << i << std::endl;
            if ((loc[i] - loc[i - 1]) == 0.)
                std::cout << "Points " << i - 1 << " and " << i
                          << " are at the same position " << loc[i]
                          << ", resulting in an infinite gradient." << std::endl;
            if (std::isnan(deltaStar[i]))
                std::cout << "Displacement thickness is nan at position " << i
                          << std::endl;
            if (std::isnan(deltaStar[i - 1]))
                std::cout << "Displacement thickness is nan at position " << i - 1
                          << std::endl;
            throw std::runtime_error(
                "NaN detected in blowing velocity computation (see reason above)");
        }
    }
}

std::vector<std::vector<double>> BoundaryLayer::evalGradwrtDeltaStar()
{
    std::vector<std::vector<double>> gradVe =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe)
        vec.resize(nNodes, 0.);

    for (size_t i = 1; i < nNodes; ++i)
    {
        gradVe[i - 1][i] = vt[i] / (loc[i] - loc[i - 1]);
        gradVe[i - 1][i - 1] =
            -1 / rhoe[i] * rhoe[i - 1] * vt[i - 1] / (loc[i] - loc[i - 1]);
    }
    return gradVe;
}

std::vector<std::vector<double>> BoundaryLayer::evalGradwrtVt()
{
    std::vector<std::vector<double>> gradVe =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe)
        vec.resize(nNodes, 0.);

    for (size_t i = 1; i < nNodes; ++i)
    {
        gradVe[i - 1][i] = deltaStar[i] / (loc[i] - loc[i - 1]);
        gradVe[i - 1][i - 1] =
            -rhoe[i - 1] / rhoe[i] * deltaStar[i - 1] / (loc[i] - loc[i - 1]);
    }
    return gradVe;
}

std::vector<std::vector<double>> BoundaryLayer::evalGradwrtRho()
{
    std::vector<std::vector<double>> gradVe =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe)
        vec.resize(nNodes, 0.);

    for (size_t i = 1; i < nNodes; ++i)
    {
        gradVe[i - 1][i] =
            -1 / (rhoe[i] * rhoe[i]) *
                ((rhoe[i] * vt[i] * deltaStar[i] -
                  rhoe[i - 1] * vt[i - 1] * deltaStar[i - 1]) /
                 (loc[i] - loc[i - 1])) +
            1 / rhoe[i] * vt[i] * deltaStar[i] / (loc[i] - loc[i - 1]);
        gradVe[i - 1][i - 1] =
            -vt[i - 1] * deltaStar[i - 1] / (rhoe[i] * (loc[i] - loc[i - 1]));
    }
    return gradVe;
}

/**
 * @brief Compute the gradient of the velocity with respect to the x coordinate.
 *
 * dVe_dx = dVe_dloc * dloc_dx
 *
 * @return std::vector<double> Gradient of the velocity with respect to the x
 * coordinate.
 */
std::vector<std::vector<double>> BoundaryLayer::evalGradwrtX()
{
    std::vector<std::vector<double>> gradVe =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe)
        vec.resize(nNodes, 0.);

    // Compute dVe/dloc
    std::vector<std::vector<double>> gradVe_loc =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe_loc)
        vec.resize(nNodes, 0.);

    for (size_t i = 1; i < nNodes; ++i)
    {
        double val = 1 / rhoe[i] *
                     (rhoe[i] * vt[i] * deltaStar[i] -
                      rhoe[i - 1] * vt[i - 1] * deltaStar[i - 1]) /
                     ((loc[i] - loc[i - 1]) * (loc[i] - loc[i - 1]));
        gradVe_loc[i - 1][i] = -val;
        gradVe_loc[i - 1][i - 1] = val;
    }

    // Compute dloc/dx
    // Derivative wrt x_j = (x_j - x_j-1) / sqrt((x_j - x_j-1)^2 + (y_j -
    // y_j-1)^2) Derivative wrt x_j-1 = - (x_j - x_j-1) / sqrt((x_j - x_j-1)^2 +
    // (y_j - y_j-1)^2) = - dloc/dx_j For loc_i, there are two contributions of
    // x_j except if j = i. loc_n = sqrt((xn-xn-1)^2) + sqrt((xn-1-xn-2)^2) + ...
    // + sqrt((x1-x0)^2)
    std::vector<std::vector<double>> gradloc_x =
        std::vector<std::vector<double>>(nNodes);
    for (auto &vec : gradloc_x)
        vec.resize(nNodes, 0.);

    for (size_t j = 1; j < nNodes; ++j)
    {
        gradloc_x[j][j] =
            (x[j] - x[j - 1]) / std::sqrt((x[j] - x[j - 1]) * (x[j] - x[j - 1]) +
                                          (y[j] - y[j - 1]) * (y[j] - y[j - 1]));
        for (size_t i = 0; i < nNodes - j; ++i)
            gradloc_x[j + i][j - 1] = gradloc_x[j - 1][j - 1] - gradloc_x[j][j];
    }

    // Matrix product of gradVe_loc and gradloc_x
    for (size_t i = 0; i < nElms; ++i)
    {
        for (size_t j = 0; j < nNodes; ++j)
        {
            for (size_t k = 0; k < nNodes; ++k)
                gradVe[i][j] += gradVe_loc[i][k] * gradloc_x[k][j];
        }
    }
    return gradVe;
}

std::vector<std::vector<double>> BoundaryLayer::evalGradwrtY()
{
    std::vector<std::vector<double>> gradVe =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe)
        vec.resize(nNodes, 0.);

    // Compute dVe/dloc
    std::vector<std::vector<double>> gradVe_loc =
        std::vector<std::vector<double>>(nElms);
    for (auto &vec : gradVe_loc)
        vec.resize(nNodes, 0.);

    for (size_t i = 1; i < nNodes; ++i)
    {
        double val = 1 / rhoe[i] *
                     (rhoe[i] * vt[i] * deltaStar[i] -
                      rhoe[i - 1] * vt[i - 1] * deltaStar[i - 1]) /
                     ((loc[i] - loc[i - 1]) * (loc[i] - loc[i - 1]));
        gradVe_loc[i - 1][i] = -val;
        gradVe_loc[i - 1][i - 1] = val;
    }

    // Compute dloc/dy
    // Derivative wrt y_j = (y_j - y_j-1) / sqrt((y_j - y_j-1)^2 + (y_j -
    // y_j-1)^2) Derivative wrt y_j-1 = - (y_j - y_j-1) / sqrt((y_j - y_j-1)^2 +
    // (y_j - y_j-1)^2) = - dloc/dy_j For loc_i, there are two contributions of
    // y_j except if j = i. loc_n = sqrt((yn-yn-1)^2) + sqrt((yn-1-yn-2)^2) + ...
    // + sqrt((y1-y0)^2)
    std::vector<std::vector<double>> gradloc_y =
        std::vector<std::vector<double>>(nNodes);
    for (auto &vec : gradloc_y)
        vec.resize(nNodes, 0.);

    for (size_t j = 1; j < nNodes; ++j)
    {
        gradloc_y[j][j] =
            (y[j] - y[j - 1]) / std::sqrt((x[j] - x[j - 1]) * (x[j] - x[j - 1]) +
                                          (y[j] - y[j - 1]) * (y[j] - y[j - 1]));
        for (size_t i = 0; i < nNodes - j; ++i)
            gradloc_y[j + i][j - 1] = gradloc_y[j - 1][j - 1] - gradloc_y[j][j];
    }

    // Matrix product of gradVe_loc and gradloc_x
    for (size_t i = 0; i < nElms; ++i)
    {
        for (size_t j = 0; j < nNodes; ++j)
        {
            for (size_t k = 0; k < nNodes; ++k)
                gradVe[i][j] += gradVe_loc[i][k] * gradloc_y[k][j];
        }
    }
    return gradVe;
}

/**
 * @brief Print solution at a given station.
 *
 * @param iPoint Node id.
 * @note Function used for debugging.
 */
void BoundaryLayer::printSolution(size_t iPoint) const
{
    if (iPoint < 0 || iPoint > nNodes - 1)
        throw std::runtime_error("Tried to access element outside of region size.");
    std::cout << "Pt " << iPoint << "Reg " << name << " Turb " << regime[iPoint]
              << std::endl;
    std::cout << std::setprecision(5) << "x " << x[iPoint] << ", xoc "
              << xoc[iPoint] << ", xx " << loc[iPoint] << "xxExt "
              << xxExt[iPoint] << std::endl;
    std::cout << std::scientific << std::setprecision(15);
    std::cout << std::setw(10) << "T " << u[iPoint * nVar + 0] << "\n"
              << std::setw(10) << "H " << u[iPoint * nVar + 1] << "\n"
              << std::setw(10) << "N " << u[iPoint * nVar + 2] << "\n"
              << std::setw(10) << "ue " << u[iPoint * nVar + 3] << "\n"
              << std::setw(10) << "Ct " << u[iPoint * nVar + 4] << "\n"
              << std::setw(10) << "dStar (BL) " << deltaStar[iPoint] << "\n"
              << std::setw(10) << "dStar (ext) " << deltaStarExt[iPoint] << "\n"
              << std::setw(10) << "vt " << vt[iPoint] << "\n"
              << std::setw(10) << "Me " << Me[iPoint] << "\n"
              << std::setw(10) << "rhoe " << rhoe[iPoint] << std::endl;
}