#include "blClosures.h"
#include <cmath>
#include <iostream>

using namespace blast;

/**
 * @brief Laminar closure relations at a given station.
 *
 * @param closureVars Vector where the computed variables will be stored.
 * @param theta Momentum thickness.
 * @param H Shape factor of the boundary layer.
 * @param Ret Reynolds number based on the momentum thickness.
 * @param Me Local Mach number.
 * @param name Name of the region.
 */
void Closures::laminarClosures(std::vector<double> &closureVars, double theta,
                               double H, double Ret, const double Me,
                               const std::string name)
{
    H = std::max(H, 1.00005);

    // Kinematic shape factor H*.
    double Hk = (H - 0.29 * Me * Me) / (1 + 0.113 * Me * Me);
    if (name == "wake")
        Hk = std::max(Hk, 1.00005);
    else
        Hk = std::max(Hk, 1.05000);

    // Density shape parameter.
    double Hstar2 = (0.064 / (Hk - 0.8) + 0.251) * Me * Me;

    // Boundary layer thickness.
    double delta = std::min((3.15 + H + (1.72 / (Hk - 1))) * theta, 12 * theta);

    double Hstar = 0.;
    double Hks = Hk - 4.35;
    if (Hk <= 4.35)
        Hstar = 0.0111 * Hks * Hks / (Hk + 1) -
                0.0278 * Hks * Hks * Hks / (Hk + 1) + 1.528 -
                0.0002 * (Hks * Hk) * (Hks * Hk);
    else
        Hstar = 1.528 + 0.015 * Hks * Hks / Hk;

    // Whitfield's minor additional compressibility correction.
    Hstar = (Hstar + 0.028 * Me * Me) / (1 + 0.014 * Me * Me);

    // Friction coefficient.
    double tmp = 0.;
    double cf = 0.;
    if (Hk < 5.5)
    {
        tmp = (5.5 - Hk) * (5.5 - Hk) * (5.5 - Hk) / (Hk + 1.0);
        cf = (-0.07 + 0.0727 * tmp) / Ret;
    }
    else if (Hk >= 5.5)
    {
        tmp = 1.0 - 1.0 / (Hk - 4.5);
        cf = (-0.07 + 0.015 * tmp * tmp) / Ret;
    }

    // Dissipation coefficient.
    double Cd = 0.;
    if (Hk < 4)
        Cd = (0.00205 * std::pow(4.0 - Hk, 5.5) + 0.207) * (Hstar / (2 * Ret));
    else
    {
        double HkCd = (Hk - 4) * (Hk - 4);
        double denCd = 1 + 0.02 * HkCd;
        Cd = (-0.0016 * HkCd / denCd + 0.207) * (Hstar / (2 * Ret));
    }

    // Wake relations.
    if (name == "wake")
    {
        Cd = 2 * (1.10 * (1.0 - 1.0 / Hk) * (1.0 - 1.0 / Hk) / Hk) *
             (Hstar / (2 * Ret));
        cf = 0.;
    }

    double us = 0.;
    double ctEq = 0.;

    closureVars = {Hstar, Hstar2, Hk, cf, Cd, delta, ctEq, us};
}

/**
 * @brief Turbulent closure relations at a given station.
 *
 * @param closureVars Vector where the computed variables will be stored.
 * @param theta Momentum thickness.
 * @param H Shape factor of the boundary layer.
 * @param Ct Shear stress coefficient.
 * @param Ret Reynolds number based on the momentum thickness.
 * @param Me Local Mach number.
 * @param name Name of the region.
 */
void Closures::turbulentClosures(std::vector<double> &closureVars, double theta,
                                 double H, double Ct, double Ret,
                                 const double Me, const std::string name)
{
    H = std::max(H, 1.00005);
    Ct = std::min(Ct, 0.3);
    // Ct = std::max(std::min(Ct, 0.30), 0.0000001);

    double Hk = (H - 0.29 * Me * Me) / (1. + 0.113 * Me * Me);
    if (name == "wake")
        Hk = std::max(Hk, 1.00005);
    else
        Hk = std::max(Hk, 1.05000);

    double Hstar2 = ((0.064 / (Hk - 0.8)) + 0.251) * Me * Me;
    double gamma = 1.4 - 1.;
    double Fc = std::sqrt(1. + 0.5 * gamma * Me * Me);

    double H0 = 0.;
    if (Ret <= 400.)
        H0 = 4.0;
    else
        H0 = 3. + (400. / Ret);
    if (Ret <= 200.)
        Ret = 200.;

    double Hstar = 0.;
    if (Hk <= H0)
        Hstar =
            ((0.5 - 4. / Ret) * ((H0 - Hk) / (H0 - 1.)) * ((H0 - Hk) / (H0 - 1.))) *
                (1.5 / (Hk + 0.5)) +
            1.5 + 4. / Ret;
    else
        Hstar = (Hk - H0) * (Hk - H0) *
                    (0.007 * std::log(Ret) /
                         ((Hk - H0 + 4. / std::log(Ret)) *
                          (Hk - H0 + 4. / std::log(Ret))) +
                     0.015 / Hk) +
                1.5 + 4. / Ret;

    // Whitfield's minor additional compressibility correction.
    Hstar = (Hstar + 0.028 * Me * Me) / (1. + 0.014 * Me * Me);

    double logRt = std::log(Ret / Fc);
    logRt = std::max(logRt, 3.0);
    double arg = std::max(-1.33 * Hk, -20.0);

    // Equivalent normalized wall slip velocity.
    double us = (Hstar / 2.) * (1. - 4. * (Hk - 1.) / (3. * H));

    // Boundary layer thickness.
    double delta = std::min((3.15 + H + (1.72 / (Hk - 1.))) * theta, 12. * theta);

    double Ctcon = 0.5 / (6.7 * 6.7 * 0.75);
    double Hkc = 0.;
    double Cdw = 0.;
    double Cdd = 0.;
    double Cdl = 0.;

    double Hmin = 0.;
    double Fl = 0.;
    double Dfac = 0.;

    double cf = 0.;
    double Cd = 0.;

    double n = 1.0;

    double ctEq = 0.;

    if (name == "wake")
    {
        if (us > 0.99995)
            us = 0.99995;
        n = 2.0;  // two wake halves
        cf = 0.0; // no friction inside the wake
        Hkc = Hk - 1.;
        Cdw = 0.0;                    // Wall contribution.
        Cdd = (0.995 - us) * Ct * Ct; // Turbulent outer layer contribution.
        Cdl = 0.15 * (0.995 - us) * (0.995 - us) /
              Ret; // Laminar stress contribution to outer layer.
        ctEq = std::sqrt(4. * Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) /
                         ((1. - us) * (Hk * Hk) * H)); // Here it is ctEq^0.5.
    }
    else
    {
        if (us > 0.95)
            us = 0.98;
        n = 1.0;
        cf = (1 / Fc) *
             (0.3 * std::exp(arg) * std::pow(logRt / 2.3026, (-1.74 - 0.31 * Hk)) +
              0.00011 * (std::tanh(4. - (Hk / 0.875)) - 1.));
        Hkc = std::max(Hk - 1. - 18. / Ret, 0.01);
        // Dissipation coefficient.
        Hmin = 1. + 2.1 / std::log(Ret);
        Fl = (Hk - 1.) / (Hmin - 1);
        Dfac = 0.5 + 0.5 * std::tanh(Fl);
        Cdw = 0.5 * (cf * us) * Dfac;
        Cdd = (0.995 - us) * Ct * Ct;
        Cdl = 0.15 * (0.995 - us) * (0.995 - us) / Ret;
        ctEq = std::sqrt(Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) /
                         ((1. - us) * (Hk * Hk) * H)); // Here it is ctEq^0.5
                                                       // ctEq = std::sqrt(Hstar * 0.015/(1-us) * (Hk-1)*(Hk-1)*(Hk-1)/(Hk*Hk*H));
                                                       // // Drela 1987
    }
    if (n * Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) /
            ((1. - us) * (Hk * Hk) * H) <
        0.)
        std::cout << "Negative sqrt encountered " << std::endl;

    // Dissipation coefficient.
    Cd = n * (Cdw + Cdd + Cdl);
    closureVars = {Hstar, Hstar2, Hk, cf, Cd, delta, ctEq, us};
}

/**
 * @brief Turbulent closure relations at a given station.
 *
 * @param closureVars Computed equilibrium shear stress coefficient.
 * @param theta Momentum thickness.
 * @param H Shape factor of the boundary layer.
 * @param Ct Shear stress coefficient.
 * @param Ret Reynolds number based on the momentum thickness.
 * @param Me Local Mach number.
 * @param name Name of the region.
 */
void Closures::turbulentClosures(double &closureVars, double theta, double H,
                                 double Ct, double Ret, const double Me,
                                 const std::string name)
{
    H = std::max(H, 1.00005);
    Ct = std::min(Ct, 0.3);

    double Hk = (H - 0.29 * Me * Me) / (1 + 0.113 * Me * Me);
    if (name == "wake")
        Hk = std::max(Hk, 1.00005);
    else
        Hk = std::max(Hk, 1.05000);

    double H0 = 0.;
    if (Ret <= 400.)
        H0 = 4.0;
    else
        H0 = 3. + (400. / Ret);
    if (Ret <= 200.)
        Ret = 200.;

    double Hstar = 0.;
    if (Hk <= H0)
        Hstar =
            ((0.5 - 4. / Ret) * ((H0 - Hk) / (H0 - 1.)) * ((H0 - Hk) / (H0 - 1.))) *
                (1.5 / (Hk + 0.5)) +
            1.5 + 4. / Ret;
    else
        Hstar = (Hk - H0) * (Hk - H0) *
                    (0.007 * std::log(Ret) /
                         ((Hk - H0 + 4. / std::log(Ret)) *
                          (Hk - H0 + 4. / std::log(Ret))) +
                     0.015 / Hk) +
                1.5 + 4. / Ret;

    // Whitfield's minor additional compressibility correction.
    Hstar = (Hstar + 0.028 * Me * Me) / (1. + 0.014 * Me * Me);

    // Equivalent normalized wall slip velocity.
    double us = (Hstar / 2.) * (1 - 4 * (Hk - 1.) / (3. * H));

    double Ctcon = 0.5 / (6.7 * 6.7 * 0.75);

    double Hkc = 0.;
    double ctEq = 0.;

    if (name == "wake")
    {
        if (us > 0.99995)
            us = 0.99995;
        Hkc = Hk - 1.;
        ctEq = std::sqrt(4. * Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) /
                         ((1. - us) * (Hk * Hk) * H)); // Here it is ctEq^0.5
    }
    else
    {
        if (us > 0.95)
            us = 0.98;
        Hkc = std::max(Hk - 1. - 18. / Ret, 0.01);
        ctEq = std::sqrt(Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) /
                         ((1. - us) * (Hk * Hk) * H)); // Here it is ctEq^0.5
    }
    if (Hstar * Ctcon * ((Hk - 1.) * Hkc * Hkc) / ((1. - us) * (Hk * Hk) * H) <
        0.)
        std::cout << "Negative sqrt encountered " << std::endl;

    closureVars = ctEq;
}