/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef BLCOUPLEDADJOINT_H
#define BLCOUPLEDADJOINT_H

#include "blast.h"
#include "dart.h"
#include "wObject.h"
#include "wTimers.h"
#include <Eigen/Sparse>

#include <iostream>
#include <memory>

namespace blast
{

/**
 * @brief Adjoint of the coupled system.
 * @author Paul Dechamps
 */
class BLAST_API CoupledAdjoint : public fwk::wSharedObject
{

public:
    std::shared_ptr<dart::Adjoint> iadjoint;     ///< Adjoint solver
    std::shared_ptr<dart::Newton> isol;          ///< Inviscid Newton solver
    std::shared_ptr<blast::Driver> vsol;         ///< Viscous solver
    std::shared_ptr<tbox::LinearSolver> alinsol; ///< linear solver (adjoint aero)

    double tdCl_AoA; ///< Total derivative of the lift coefficient with respect to
                     ///< the angle of attack
    double tdCd_AoA; ///< Total erivative of the drag coefficient with respect to
                     ///< the angle of attack

    std::vector<Eigen::Vector3d>
        tdCl_x; ///< Total derivative of the lift coefficient with respect to the
                ///< mesh coordinates
    std::vector<Eigen::Vector3d>
        tdCd_x; ///< Total derivative of the drag coefficient with respect to the
                ///< mesh coordinates

private:
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_phi; ///< Inviscid flow Jacobian
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_M; ///< Derivative of the inviscid flow residual with respect to the
                 ///< Mach number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_v; ///< Derivative of the inviscid flow residual with respect to the
                 ///< velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_rho; ///< Derivative of the inviscid flow residual with respect to
                   ///< the density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_dStar; ///< Derivative of the inviscid flow residual with respect to
                     ///< delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_blw; ///< Derivative of the inviscid flow residual with respect to
                   ///< the blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRphi_x; ///< Derivative of the potential residual with respect to the
                 ///< mesh coordinates

    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_phi; ///< Derivative of the Mach number residual with respect to the
                 ///< inviscid flow
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_M; ///< Derivative of the Mach number residual with respect to the
               ///< Mach number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_v; ///< Derivative of the Mach number residual with respect to the
               ///< velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_rho; ///< Derivative of the Mach number residual with respect to the
                 ///< density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_dStar; ///< Derivative of the Mach number residual with respect to
                   ///< delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_blw; ///< Derivative of the Mach number residual with respect to the
                 ///< blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRM_x; ///< Derivative of the Mach number residual with respect to the
               ///< mesh coordinates

    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_phi; ///< Derivative of the density residual with respect to the
                   ///< inviscid flow
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_M; ///< Derivative of the density residual with respect to the Mach
                 ///< number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_v; ///< Derivative of the density residual with respect to the
                 ///< velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_rho; ///< Derivative of the density residual with respect to the
                   ///< density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_dStar; ///< Derivative of the density residual with respect to
                     ///< delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_blw; ///< Derivative of the density residual with respect to the
                   ///< blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRrho_x; ///< Derivative of the density residual with respect to the mesh
                 ///< coordinates

    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_phi; ///< Derivative of the velocity residual with respect to the
                 ///< inviscid flow
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_M; ///< Derivative of the velocity residual with respect to the Mach
               ///< number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_v; ///< Derivative of the velocity residual with respect to the
               ///< velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_rho; ///< Derivative of the velocity residual with respect to the
                 ///< density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_dStar; ///< Derivative of the velocity residual with respect to delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_blw; ///< Derivative of the velocity residual with respect to the
                 ///< blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRv_x; ///< Derivative of the velocity residual with respect to the mesh
               ///< coordinates

    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_phi; ///< Derivative of delta* residual with respect to the
                     ///< inviscid flow
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_M; ///< Derivative of delta* residual with respect to the Mach
                   ///< number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_v; ///< Derivative of delta* residual with respect to the velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_rho; ///< Derivative of delta* residual with respect to the
                     ///< density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_dStar; ///< Derivative of delta* residual with respect to delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_blw; ///< Derivative of delta* residual with respect to the
                     ///< blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRdStar_x; ///< Derivative of delta* residual with respect to the mesh
                   ///< coordinates

    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_phi; ///< Derivative of the blowing velocity residual with respect
                   ///< to the inviscid flow
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_M; ///< Derivative of the blowing velocity residual with respect to
                 ///< the Mach number
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_v; ///< Derivative of the blowing velocity residual with respect to
                 ///< the velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_rho; ///< Derivative of the blowing velocity residual with respect
                   ///< to the density
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_dStar; ///< Derivative of the blowing velocity residual with respect
                     ///< to delta*
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_blw; ///< Derivative of the blowing velocity residual with respect
                   ///< to the blowing velocity
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRblw_x; ///< Derivative of the blowing velocity residual with respect to
                 ///< the mesh coordinates

    // Objective functions
    Eigen::VectorXd dCl_phi; ///< Derivative of the lift coefficient with respect
                             ///< to the inviscid flow
    Eigen::VectorXd dCd_phi; ///< Derivative of the drag coefficient with respect
                             ///< to the inviscid flow
    Eigen::VectorXd dCd_M;   ///< Derivative of the drag coefficient with respect to
                             ///< the Mach number
    Eigen::VectorXd dCd_rho; ///< Derivative of the drag coefficient with respect
                             ///< to the density
    Eigen::VectorXd dCd_v;   ///< Derivative of the drag coefficient with respect to
                             ///< the velocity
    Eigen::VectorXd
        dCd_dStar; ///< Derivative of the drag coefficient with respect to delta*

    // Angle of attack
    Eigen::VectorXd dRphi_AoA; ///< Derivative of the inviscid flow residual with
                               ///< respect to the angle of attack
    double dCl_AoA;            ///< Partial derivative of the lift coefficient with respect
                               ///< to the angle of attack
    double dCd_AoA;            ///< Partial derivative of the total drag coefficient with
                               ///< respect to the angle of attack

    // Mesh
    Eigen::SparseMatrix<double, Eigen::RowMajor>
        dRx_x;              ///< Derivative of the mesh residual with respect to the mesh
                            ///< coordinates
    Eigen::VectorXd dCl_x;  ///< Partial derivative of the lift coefficient with
                            ///< respect to the mesh coordinates
    Eigen::VectorXd dCdp_x; ///< Partial derivative of the pressure drag
                            ///< coefficient with respect to the mesh coordinates
    Eigen::VectorXd dCdf_x; ///< Partial derivative of the friction drag
                            ///< coefficient with respect to the mesh coordinates

    fwk::Timers tms;  //< internal timers
    fwk::Timers tms2; //< internal precise timers

public:
    CoupledAdjoint(std::shared_ptr<dart::Adjoint> iAdjoint,
                   std::shared_ptr<blast::Driver> vSolver);
    virtual ~CoupledAdjoint() { std::cout << "~blast::CoupledAdjoint()\n"; }

    void run();
    void reset();

private:
    void buildAdjointMatrix(
        std::vector<std::vector<Eigen::SparseMatrix<double, Eigen::RowMajor> *>>
            &matrices,
        Eigen::SparseMatrix<double, Eigen::RowMajor> &matrixAdjoint);

    Eigen::VectorXd runViscous();

    void gradientswrtInviscidFlow();
    void gradientswrtMachNumber();
    void gradientswrtDensity();
    void gradientswrtVelocity();
    void gradientswrtDeltaStar();
    void gradientswrtBlowingVelocity();

    void gradientwrtAoA();
    void gradientwrtMesh();

    void transposeMatrices();
    void writeMatrixToFile(const Eigen::MatrixXd &matrix,
                           const std::string &filename);
};
} // namespace blast
#endif // BLCOUPLEDADJOINT_H
