#include "blDriver.h"
#include "blBoundaryLayer.h"
#include "blSolver.h"
#include <iomanip>

#define ANSI_COLOR_RED "\x1b[1;31m"
#define ANSI_COLOR_GREEN "\x1b[1;32m"
#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

using namespace blast;

/**
 * @brief Driver of the viscous solver.
 *
 * @param _Re Freestream Reynolds number.
 * @param _Minf Freestream Mach number.
 * @param _CFL0 Initial CFL number for pseudo-time integration.
 * @param nSections Number of sections in the computation (1 for 2D cases).
 * @param _Span Span of the considered wing (only used for drag computation).
 * @param _verbose Verbosity level of the solver 0<=verbose<=1.
 */
Driver::Driver(double _Re, double _Minf, double _CFL0, size_t nSections,
               double xtrF_top, double xtrF_bot, double _span,
               unsigned int _verbose)
{
    // Freestream parameters.
    Re = _Re;
    CFL0 = _CFL0;
    verbose = _verbose;

    std::vector<double> xtrF = {xtrF_top, xtrF_bot, 0.};

    // Initialzing boundary layer
    sections.resize(nSections);
    solverExitCode = 0;
    convergenceStatus.resize(nSections);
    for (size_t iSec = 0; iSec < nSections; ++iSec)
        convergenceStatus[iSec].resize(3);

    Cdt_sec.resize(sections.size(), 0.);
    Cdf_sec.resize(sections.size(), 0.);
    Cdp_sec.resize(sections.size(), 0.);
    Cdt = 0.;
    Cdf = 0.;
    Cdp = 0.;
    span = _span;

    // Time solver initialization.
    tSolver = new Solver(_CFL0, _Minf, Re);

    // Numerical parameters.
    std::cout << "--- Viscous solver setup ---\n"
              << "Reynolds number: " << Re << " \n"
              << "Freestream Mach number: " << _Minf << " \n"
              << "Initial CFL number: " << CFL0 << " \n"
              << std::endl;
    std::cout << "Boundary layer initialized." << std::endl;
}

/**
 * @brief Driver and subsequent dependencies destruction.
 */
Driver::~Driver()
{
    delete tSolver;
    std::cout << "~blast::Driver()\n";
} // end ~Driver

void Driver::reset()
{
    size_t nSections = sections.size();
    for (size_t iSec = 0; iSec < sections.size(); ++iSec)
        sections[iSec].resize(0);
    sections.resize(nSections);
}

/**
 * @brief Runs viscous operations. Temporal solver parametrisation is first
 * adapted. Solutions are initialized than time marched towards steady state
 * successively for each points.
 *
 * @return int Solver exit code.
 */
int Driver::run()
{
    bool lockTrans = false; // Flag activation of transition routines.
    int pointExitCode;      // Output of pseudo time integration (0: converged; -1:
                            // unsuccessful, failed; >0: unsuccessful nb iter).

    double maxMach = 0.;

    for (size_t iSec = 0; iSec < sections.size(); ++iSec)
    {
        if (verbose > 1)
            std::cout << "Sec " << iSec << " ";

        // Loop over the regions (upper, lower and wake).
        int iRegion = 0;
        for (auto reg : sections[iSec])
        {
            convergenceStatus[iSec][iRegion].resize(0);

            if (reg->getMaxMach() > maxMach)
                maxMach = reg->getMaxMach();

            // Reset transition
            if (reg->name == "wake")
                reg->xtr = 0.;
            else if (reg->name == "upper" || reg->name == "lower")
                reg->xtr = 1.;
            else
            {
                std::cout << "reg->name " << reg->name << std::endl;
                throw std::runtime_error("Wrong region name\n");
            }

            // Reset regime.
            lockTrans = false;
            if (reg->name == "upper" || reg->name == "lower")
            {
                for (size_t k = 0; k < reg->nNodes; k++)
                    reg->regime[k] = 0;
                reg->setStagnationBC(Re);
            }

            else if (reg->name == "wake") // Wake
            {
                for (size_t k = 0; k < reg->nNodes; k++)
                    reg->regime[k] = 1;

                // Impose wake boundary condition.
                std::vector<double> upperConditions(reg->getnVar(), 0.);
                std::vector<double> lowerConditions(reg->getnVar(), 0.);
                for (size_t k = 0; k < upperConditions.size(); ++k)
                {
                    upperConditions[k] =
                        sections[iSec][0]->u[(sections[iSec][0]->nNodes - 1) *
                                                 sections[iSec][0]->getnVar() +
                                             k];
                    lowerConditions[k] =
                        sections[iSec][1]->u[(sections[iSec][1]->nNodes - 1) *
                                                 sections[iSec][1]->getnVar() +
                                             k];
                }
                reg->setWakeBC(Re, upperConditions, lowerConditions);
                lockTrans = true;
            }
            else
                throw std::runtime_error("Wrong region name\n");

            // Loop over points
            for (size_t iPoint = 1; iPoint < reg->nNodes; ++iPoint)
            {
                // Initialize solution at point
                tSolver->initialCondition(iPoint, reg);
                // Solve equations
                tms["1-Solver"].start();
                pointExitCode = tSolver->integration(iPoint, reg);
                tms["1-Solver"].stop();

                if (pointExitCode != 0)
                    convergenceStatus[iSec][iRegion].push_back(iPoint);

                // Transition
                tms["2-Transition"].start();
                if (!lockTrans)
                {
                    // Forced transition
                    if (reg->xtrF != -1 && reg->xoc[iPoint] >= reg->xtrF)
                    {
                        reg->u[iPoint * reg->getnVar() + 2] = reg->getnCrit();
                        averageTransition(iPoint, reg, true);
                        if (verbose > 1)
                        {
                            std::cout << std::fixed;
                            std::cout << std::setprecision(2);
                            std::cout << reg->xtr << " (" << reg->transitionNode << ") ";
                        }
                        lockTrans = true;
                    }
                    // Free transtion
                    else if (reg->xtrF == -1)
                    {
                        // Check if perturbation waves are growing or not.
                        checkWaves(iPoint, reg);

                        // Amplification factor is compared to critical amplification
                        // factor.
                        if (reg->u[iPoint * reg->getnVar() + 2] >= reg->getnCrit())
                        {
                            averageTransition(iPoint, reg, false);
                            if (verbose > 1)
                            {
                                std::cout << std::fixed;
                                std::cout << std::setprecision(2);
                                std::cout << reg->xtr << " (" << reg->transitionNode << ") ";
                            }
                            lockTrans = true;
                        }
                    }
                }
                tms["2-Transition"].stop();
            }
            tms["3-Blowing"].start();
            reg->computeBlowingVelocity();
            tms["3-Blowing"].stop();
            iRegion++;
        }
        if (verbose > 1)
            std::cout << std::endl;
    }

    for (size_t i = 0; i < sections.size(); ++i)
        computeSectionalDrag(sections[i], i);
    computeTotalDrag();

    // Output information.
    solverExitCode = outputStatus(maxMach);

    return solverExitCode; // exit code
}

/**
 * @brief Check whether or not instabilities are growing in the laminar boundary
 * layer.
 *
 * @param iPoint Marker id.
 * @param bl BoundaryLayer region.
 */
void Driver::checkWaves(size_t iPoint, BoundaryLayer *bl)
{

    double logRet_crit =
        2.492 * std::pow(1. / (bl->Hk[iPoint] - 1), 0.43) +
        0.7 * (std::tanh(14 * (1. / (bl->Hk[iPoint] - 1)) - 9.24) + 1.0);

    if (bl->Ret[iPoint] > 0) // Avoid log of negative number.
    {
        if (std::log10(bl->Ret[iPoint]) <= logRet_crit - 0.08)
            bl->u[iPoint * bl->getnVar() + 2] =
                0.; // Set N to 0 at that point if waves do not grow.
    }
    else
        bl->u[iPoint * bl->getnVar() + 2] =
            0.; // Set N to 0 at that point if waves do not grow.
}

/**
 * @brief Computes turbulent solution @ the transition point and averages
 * solutions.
 *
 * @param iPoint Marker id.
 * @param bl BoundaryLayer region.
 * @param forced Flag for forced transition.
 */
void Driver::averageTransition(size_t iPoint, BoundaryLayer *bl, bool forced)
{
    // Averages solution on transition marker.
    size_t nVar = bl->getnVar();

    // Save laminar solution.
    std::vector<double> lamSol(nVar, 0.0);
    for (size_t k = 0; k < lamSol.size(); ++k)
        lamSol[k] = bl->u[iPoint * nVar + k];

    // Set up turbulent portion boundary condition.
    bl->transitionNode = iPoint; // Save transition marker.

    // Loop over remaining points in the region.
    for (size_t k = iPoint; k < bl->nNodes; ++k)
        bl->regime[k] = 1; // Set Turbulent regime.

    bl->xtr = (bl->getnCrit() - bl->u[(iPoint - 1) * nVar + 2]) *
                  ((bl->xoc[iPoint] - bl->xoc[iPoint - 1]) /
                   (bl->u[iPoint * nVar + 2] - bl->u[(iPoint - 1) * nVar + 2])) +
              bl->xoc[iPoint - 1];

    double avTurb =
        (bl->xoc[iPoint] - bl->xtr) / (bl->xoc[iPoint] - bl->xoc[iPoint - 1]);
    double avLam = 1. - avTurb;

    // std::cout << "avTurb " << avTurb << " avLam " << avLam << std::endl;
    if (avTurb < 0. || avTurb > 1. || avLam < 0. || avLam > 1.)
    {
        bl->printSolution(iPoint);
        bl->printSolution(iPoint - 1);
        std::cout << "dx " << std::abs(bl->xoc[iPoint] - bl->xoc[iPoint - 1])
                  << std::endl;
        std::cout << " (bl->u[iPoint * nVar + 2] - bl->u[(iPoint - 1) * nVar + 2]) "
                  << (bl->u[iPoint * nVar + 2] - bl->u[(iPoint - 1) * nVar + 2])
                  << std::endl;
        std::cout << " bl->getnCrit() - bl->u[(iPoint - 1) * nVar + 2] "
                  << bl->getnCrit() - bl->u[(iPoint - 1) * nVar + 2] << std::endl;
        std::cout << "xtr " << bl->xtr << std::endl;
        std::cout << "avTurb " << avTurb << " avLam " << avLam << std::endl;
        throw std::runtime_error("Transition location out of bounds.");
    }

    // Impose boundary condition.
    double Cteq_trans;
    bl->closSolver->turbulentClosures(Cteq_trans, bl->u[(iPoint - 1) * nVar + 0],
                                      bl->u[(iPoint - 1) * nVar + 1], 0.03,
                                      bl->Ret[iPoint - 1], bl->Me[iPoint - 1],
                                      bl->name);
    bl->ctEq[iPoint - 1] = Cteq_trans;
    bl->u[(iPoint - 1) * nVar + 4] = 0.7 * bl->ctEq[iPoint - 1];

    // Avoid starting with ill conditioned IC for turbulent computation. (Regime
    // was switched above). These initial guess values do not influence the
    // converged solution.
    bl->u[iPoint * nVar + 4] =
        bl->u[(iPoint - 1) * nVar + 4]; // IC of transition point = turbulent BC
                                        // (imposed @ previous point).
    bl->u[iPoint * nVar + 1] =
        1.515; // Because we expect a significant drop of the shape factor.

    // Solve point in turbulent configuration.
    int exitCode = tSolver->integration(iPoint, bl);

    if (exitCode != 0)
        std::cout << "Warning: Transition point turbulent computation terminated "
                     "with exit code "
                  << exitCode << std::endl;

    // Average both solutions (Now solution @ iPoint is the turbulent solution).
    bl->u[iPoint * nVar + 0] =
        avLam * lamSol[0] + avTurb * bl->u[(iPoint)*nVar + 0]; // Theta.
    bl->u[iPoint * nVar + 1] =
        avLam * lamSol[1] + avTurb * bl->u[(iPoint)*nVar + 1]; // H.
    bl->u[iPoint * nVar + 2] = bl->getnCrit();                 // N.
    bl->u[iPoint * nVar + 3] =
        avLam * lamSol[3] + avTurb * bl->u[(iPoint)*nVar + 3];    // ue.
    bl->u[iPoint * nVar + 4] = avTurb * bl->u[(iPoint)*nVar + 4]; // Ct.

    bl->deltaStar[iPoint - 1] =
        bl->u[(iPoint - 1) * nVar + 0] * bl->u[(iPoint - 1) * nVar + 1];
    bl->deltaStar[iPoint] = bl->u[iPoint * nVar + 0] * bl->u[iPoint * nVar + 1];

    // Recompute closures. (The turbulent BC @ iPoint - 1 does not influence
    // laminar closure relations).
    std::vector<double> lamParam(8, 0.);
    bl->closSolver->laminarClosures(
        lamParam, bl->u[(iPoint - 1) * nVar + 0], bl->u[(iPoint - 1) * nVar + 1],
        bl->Ret[iPoint - 1], bl->Me[iPoint - 1], bl->name);
    bl->Hstar[iPoint - 1] = lamParam[0];
    bl->Hstar2[iPoint - 1] = lamParam[1];
    bl->Hk[iPoint - 1] = lamParam[2];
    bl->cf[iPoint - 1] = lamParam[3];
    bl->cd[iPoint - 1] = lamParam[4];
    bl->delta[iPoint - 1] = lamParam[5];
    bl->ctEq[iPoint - 1] = lamParam[6];
    bl->us[iPoint - 1] = lamParam[7];

    std::vector<double> turbParam(8, 0.);
    bl->closSolver->turbulentClosures(
        turbParam, bl->u[(iPoint)*nVar + 0], bl->u[(iPoint)*nVar + 1],
        bl->u[(iPoint)*nVar + 4], bl->Ret[iPoint], bl->Me[iPoint], bl->name);
    bl->Hstar[iPoint] = turbParam[0];
    bl->Hstar2[iPoint] = turbParam[1];
    bl->Hk[iPoint] = turbParam[2];
    bl->cf[iPoint] = turbParam[3];
    bl->cd[iPoint] = turbParam[4];
    bl->delta[iPoint] = turbParam[5];
    bl->ctEq[iPoint] = turbParam[6];
    bl->us[iPoint] = turbParam[7];
}

/**
 * @brief Friction, pressure and total drag computation at each station of the
 * configuration.
 *
 * @tparam Cdt = theta * Ue^((H+5)/2).
 * @tparam Cdf = integral(Cf dx)_upper + integral(Cf dx)_lower.
 * @tparam Cdp = Cdtot - Cdf.
 *
 * @param bl BoundaryLayer region.
 * @param i Considered section.
 */
void Driver::computeSectionalDrag(std::vector<BoundaryLayer *> const &sec,
                                  size_t i)
{
    size_t nVar = sec[0]->getnVar();

    // Normalize friction and dissipation coefficients.
    std::vector<double> frictioncoeff_upper(sec[0]->nNodes, 0.0);
    for (size_t k = 0; k < frictioncoeff_upper.size(); ++k)
        frictioncoeff_upper[k] = sec[0]->vt[k] * sec[0]->vt[k] * sec[0]->cf[k];

    std::vector<double> frictioncoeff_lower(sec[1]->nNodes, 0.0);
    for (size_t k = 0; k < frictioncoeff_lower.size(); ++k)
        frictioncoeff_lower[k] = sec[1]->vt[k] * sec[1]->vt[k] * sec[1]->cf[k];

    // Compute friction drag coefficient.
    double Cdf_upper = 0.0;
    for (size_t k = 1; k < sec[0]->nNodes; ++k)
        Cdf_upper += (frictioncoeff_upper[k] + frictioncoeff_upper[k - 1]) *
                     (sec[0]->x[k] - sec[0]->x[k - 1]) / 2;
    double Cdf_lower = 0.0;
    for (size_t k = 1; k < sec[1]->nNodes; ++k)
        Cdf_lower += (frictioncoeff_lower[k] + frictioncoeff_lower[k - 1]) *
                     (sec[1]->x[k] - sec[1]->x[k - 1]) / 2;
    Cdf_sec[i] = Cdf_upper + Cdf_lower; // No friction in the wake

    // Total drag coefficient. Computed from upper and lower TE if there is no
    // wake.
    if (sec.size() > 2 && sec[2]->getName() == "wake")
    {
        size_t lastWkPt = (sec[2]->nNodes - 1) * nVar;
        Cdt_sec[i] =
            2 * sec[2]->u[lastWkPt + 0] *
            pow(sec[2]->u[lastWkPt + 3], (sec[2]->u[lastWkPt + 1] + 5) / 2);
    }
    else
    {
        size_t lastUpPt = (sec[0]->nNodes - 1) * nVar;
        size_t lastLwPt = (sec[1]->nNodes - 1) * nVar;
        Cdt_sec[i] =
            sec[0]->u[lastUpPt + 0] *
                pow(sec[0]->u[lastUpPt + 3], (sec[0]->u[lastUpPt + 1] + 5) / 2) +
            sec[1]->u[lastLwPt + 0] *
                pow(sec[1]->u[lastLwPt + 3], (sec[1]->u[lastLwPt + 1] + 5) / 2);
    }
    // Pressure drag coefficient.
    Cdp_sec[i] = Cdt_sec[i] - Cdf_sec[i];
}

double Driver::getAverageTransition(size_t const iRegion) const
{
    double xtr = 0.0;
    for (size_t iSec = 0; iSec < sections.size(); ++iSec)
        xtr += sections[iSec][iRegion]->xtr;
    xtr /= sections.size();
    return xtr;
}

/**
 * @brief Compute total drag coefficient on the airfoil/wing.
 * Performs the sectional drag integration for 3D computations.
 *
 */
void Driver::computeTotalDrag()
{
    Cdt = 0.;
    Cdf = 0.;
    Cdp = 0.;

    if (sections.size() == 1 || span == 0.)
    {
        Cdt = Cdt_sec[0];
        Cdf = Cdf_sec[0];
        Cdp = Cdp_sec[0];
    }
    else
    {
        Cdt += (sections[0][0]->z[0] - 0) * Cdt_sec[0];
        Cdp += (sections[0][0]->z[0] - 0) * Cdp_sec[0];
        Cdf += (sections[0][0]->z[0] - 0) * Cdf_sec[0];
        double dz = 0.;
        for (size_t k = 1; k < sections.size(); ++k)
        {
            dz = (sections[k][0]->z[0] - sections[k - 1][0]->z[0]);
            Cdt += dz * Cdt_sec[k];
            Cdp += dz * Cdp_sec[k];
            Cdf += dz * Cdf_sec[k];
        }
        Cdt /= span;
        Cdp /= span;
        Cdf /= span;
    }
}

/**
 * @brief Outputs solver status depending on the verbosity level (verbose).
 *
 * @param maxMach Maximum Mach number identified on the configuration.
 * @return int Solver exit code (0 converged, !=0 not converged).
 */
int Driver::outputStatus(double maxMach) const
{
    int solverExitCode = 0;
    for (size_t iSec = 0; iSec < sections.size(); ++iSec)
        for (size_t iReg = 0; iReg < sections[iSec].size(); ++iReg)
            if (convergenceStatus[iSec][iReg].size() != 0)
            {
                solverExitCode = -1;
                break;
            }

    if (verbose > 2 && solverExitCode != 0)
    {
        // Output unconverged points.
        std::cout << "Unconverged points" << std::endl;
        std::cout << std::setw(10) << "Section" << std::setw(12) << "Region"
                  << std::setw(16) << "Markers" << std::endl;
        for (size_t iSec = 0; iSec < sections.size(); ++iSec)
        {
            unsigned int count = 0;
            for (size_t iReg = 0; iReg < sections[iSec].size(); ++iReg)
            {
                std::cout << std::setw(10) << iSec;
                std::cout << std::setw(12) << iReg << "          ";
                if (convergenceStatus[iSec][iReg].size() != 0)
                    for (size_t iPoint = 0; iPoint < convergenceStatus[iSec][iReg].size();
                         ++iPoint)
                    {
                        std::cout << convergenceStatus[iSec][iReg][iPoint] << " ("
                                  << sections[iSec][iReg]
                                         ->x[convergenceStatus[iSec][iReg][iPoint]]
                                  << "), ";
                        ++count;
                    }
                std::cout << "\n";
            }
            std::cout << "total: " << count << "\n";
        }
        std::cout << "" << std::endl;
    }

    if (verbose > 0)
    {
        // Compute mean transition locations.
        std::vector<double> meanTransitions(2, 0.);
        for (size_t iSec = 0; iSec < sections.size(); ++iSec)
            for (size_t iReg = 0; iReg < 2; ++iReg)
                meanTransitions[iReg] += sections[iSec][iReg]->xtr;
        for (size_t iReg = 0; iReg < meanTransitions.size(); ++iReg)
            meanTransitions[iReg] /= sections.size();

        std::cout << std::fixed << std::setprecision(2) << "Viscous solver for Re "
                  << Re / 1e6 << "e6, Mach max " << maxMach << std::endl;

        std::cout << std::setw(10) << "xTr Upper" << std::setw(12) << "xTr Lower"
                  << std::setw(12) << "Cd" << std::endl;
        std::cout << std::fixed << std::setprecision(2);
        std::cout << std::setw(10) << meanTransitions[0] << std::setw(12)
                  << meanTransitions[1];
        std::cout << std::fixed << std::setprecision(5);
        std::cout << std::setw(12) << Cdt << "\n";

        if (verbose > 0)
        {
            if (solverExitCode == 0)
                std::cout << ANSI_COLOR_GREEN << "Viscous solver converged."
                          << ANSI_COLOR_RESET << std::endl;
            else
                std::cout << ANSI_COLOR_YELLOW << "Viscous solver not converged."
                          << ANSI_COLOR_RESET << std::endl;
        }

        if (verbose > 2)
        {
            std::cout << "Viscous solver CPU" << std::endl
                      << tms;
        }
    }

    if (verbose > 0)
        std::cout << "" << std::endl;
    return solverExitCode;
}

/**
 * @brief Returns the integral boundary layer solution in a section (sorted from
 * upper Te to lower Te than wake).
 *
 * @param iSec id of the considered section.
 * @return std::vector<std::vector<double>>
 */
std::vector<std::vector<double>> Driver::getSolution(size_t iSec)
{
    size_t nMarkersUp = sections[iSec][0]->nNodes;
    size_t nMarkersLw = sections[iSec][1]->nNodes; // No stagnation point doublon.
    size_t nVar = sections[iSec][0]->getnVar();

    std::vector<std::vector<double>> Sln(20);
    for (size_t iVector = 0; iVector < Sln.size(); ++iVector)
        Sln[iVector].resize(nMarkersUp + nMarkersLw + sections[iSec][2]->nNodes,
                            0.);
    // Upper side.
    size_t revPoint = nMarkersUp - 1;
    for (size_t iPoint = 0; iPoint < nMarkersUp; ++iPoint)
    {
        Sln[0][iPoint] = sections[iSec][0]->u[revPoint * nVar + 0];
        Sln[1][iPoint] = sections[iSec][0]->u[revPoint * nVar + 1];
        Sln[2][iPoint] = sections[iSec][0]->u[revPoint * nVar + 2];
        Sln[3][iPoint] = sections[iSec][0]->u[revPoint * nVar + 3];
        Sln[4][iPoint] = sections[iSec][0]->u[revPoint * nVar + 4];
        Sln[5][iPoint] = sections[iSec][0]->deltaStar[revPoint];

        Sln[6][iPoint] = sections[iSec][0]->Ret[revPoint];
        Sln[7][iPoint] = sections[iSec][0]->cd[revPoint];
        Sln[8][iPoint] = sections[iSec][0]->cf[revPoint];
        Sln[9][iPoint] = sections[iSec][0]->Hstar[revPoint];
        Sln[10][iPoint] = sections[iSec][0]->Hstar2[revPoint];
        Sln[11][iPoint] = sections[iSec][0]->Hk[revPoint];
        Sln[12][iPoint] = sections[iSec][0]->ctEq[revPoint];
        Sln[13][iPoint] = sections[iSec][0]->us[revPoint];
        Sln[14][iPoint] = sections[iSec][0]->delta[revPoint];

        Sln[15][iPoint] = sections[iSec][0]->x[revPoint];
        Sln[16][iPoint] = sections[iSec][0]->xoc[revPoint];
        Sln[17][iPoint] = sections[iSec][0]->y[revPoint];
        Sln[18][iPoint] = sections[iSec][0]->z[revPoint];
        Sln[19][iPoint] = sections[iSec][0]->vt[revPoint];
        --revPoint;
    }
    // Lower side.
    for (size_t iPoint = 0; iPoint < sections[iSec][1]->nNodes; ++iPoint)
    {
        Sln[0][nMarkersUp + iPoint] = sections[iSec][1]->u[iPoint * nVar + 0];
        Sln[1][nMarkersUp + iPoint] = sections[iSec][1]->u[iPoint * nVar + 1];
        Sln[2][nMarkersUp + iPoint] = sections[iSec][1]->u[iPoint * nVar + 2];
        Sln[3][nMarkersUp + iPoint] = sections[iSec][1]->u[iPoint * nVar + 3];
        Sln[4][nMarkersUp + iPoint] = sections[iSec][1]->u[iPoint * nVar + 4];
        Sln[5][nMarkersUp + iPoint] = sections[iSec][1]->deltaStar[iPoint];

        Sln[6][nMarkersUp + iPoint] = sections[iSec][1]->Ret[iPoint];
        Sln[7][nMarkersUp + iPoint] = sections[iSec][1]->cd[iPoint];
        Sln[8][nMarkersUp + iPoint] = sections[iSec][1]->cf[iPoint];
        Sln[9][nMarkersUp + iPoint] = sections[iSec][1]->Hstar[iPoint];
        Sln[10][nMarkersUp + iPoint] = sections[iSec][1]->Hstar2[iPoint];
        Sln[11][nMarkersUp + iPoint] = sections[iSec][1]->Hk[iPoint];
        Sln[12][nMarkersUp + iPoint] = sections[iSec][1]->ctEq[iPoint];
        Sln[13][nMarkersUp + iPoint] = sections[iSec][1]->us[iPoint];
        Sln[14][nMarkersUp + iPoint] = sections[iSec][1]->delta[iPoint];

        Sln[15][nMarkersUp + iPoint] = sections[iSec][1]->x[iPoint];
        Sln[16][nMarkersUp + iPoint] = sections[iSec][1]->xoc[iPoint];
        Sln[17][nMarkersUp + iPoint] = sections[iSec][1]->y[iPoint];
        Sln[18][nMarkersUp + iPoint] = sections[iSec][1]->z[iPoint];
        Sln[19][nMarkersUp + iPoint] = sections[iSec][1]->vt[iPoint];
    }
    // Wake.
    for (size_t iPoint = 0; iPoint < sections[iSec][2]->nNodes; ++iPoint)
    {
        Sln[0][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->u[iPoint * nVar + 0];
        Sln[1][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->u[iPoint * nVar + 1];
        Sln[2][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->u[iPoint * nVar + 2];
        Sln[3][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->u[iPoint * nVar + 3];
        Sln[4][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->u[iPoint * nVar + 4];
        Sln[5][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->deltaStar[iPoint];

        Sln[6][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->Ret[iPoint];
        Sln[7][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->cd[iPoint];
        Sln[8][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->cf[iPoint];
        Sln[9][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->Hstar[iPoint];
        Sln[10][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->Hstar2[iPoint];
        Sln[11][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->Hk[iPoint];
        Sln[12][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->ctEq[iPoint];
        Sln[13][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->us[iPoint];
        Sln[14][nMarkersUp + nMarkersLw + iPoint] =
            sections[iSec][2]->delta[iPoint];

        Sln[15][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->x[iPoint];
        Sln[16][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->xoc[iPoint];
        Sln[17][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->y[iPoint];
        Sln[18][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->z[iPoint];
        Sln[19][nMarkersUp + nMarkersLw + iPoint] = sections[iSec][2]->vt[iPoint];
    }

    if (verbose > 2)
        std::cout << "Solution structure sent." << std::endl;
    return Sln;
}