#ifndef BLDRIVER_H
#define BLDRIVER_H

#include "blBoundaryLayer.h"
#include "blast.h"
#include "wObject.h"
#include "wTimers.h"
namespace blast
{

/**
 * @brief Boundary layer solver driver. Performs the space marching and set up
 * time marching for each point.
 * @authors Paul Dechamps
 */
class BLAST_API Driver : public fwk::wSharedObject
{
public:
    double Cdt;                  ///< Total drag coefficient.
    double Cdf;                  ///< Total friction drag coefficient.
    double Cdp;                  ///< Total pressure drag coefficient.
    std::vector<double> Cdt_sec; ///< Sectional total drag coefficient.
    std::vector<double> Cdf_sec; ///< Sectional friction drag coefficient.
    std::vector<double> Cdp_sec; ///< Sectional pressure drag coefficient.
    std::vector<std::vector<BoundaryLayer *>>
        sections;         ///< Boundary layer regions sorted by section.
    unsigned int verbose; ///< Verbosity level of the solver 0<=verbose<=1.

private:
    double Re;          ///< Freestream Reynolds number.
    double CFL0;        ///< Initial CFL number for pseudo time integration.
    double span;        ///< Wing Span (Used only for drag computation, not used if 2D
                        ///< case).
    Solver *tSolver;    ///< Pseudo-time solver.
    int solverExitCode; ///< Exit code of viscous calculations.
    std::vector<std::vector<std::vector<size_t>>>
        convergenceStatus; ///< Vector containing points that did not converged.

protected:
    fwk::Timers tms; ///< Internal timers

public:
    Driver(double _Re, double _Minf, double _CFL0, size_t nSections,
           double xtrF_top = -1., double xtrF_bot = -1., double _span = 0.,
           unsigned int _verbose = 1);
    ~Driver();
    int run();
    void reset();

    // Getters.
    double getAverageTransition(size_t const iRegion) const;
    double getRe() const { return Re; }
    std::vector<std::vector<double>> getSolution(size_t iSec);

    // Setters.
    void setVerbose(unsigned int _verbose) { verbose = _verbose; }

    void addSection(size_t iSec, BoundaryLayer *&bl)
    {
        sections[iSec].push_back(bl);
    }

private:
    void checkWaves(size_t iPoint, BoundaryLayer *bl);
    void averageTransition(size_t iPoint, BoundaryLayer *bl, bool forced);
    void computeSectionalDrag(std::vector<BoundaryLayer *> const &bl, size_t i);
    void computeTotalDrag();
    void computeBlwVel();
    int outputStatus(double maxMach) const;
};
} // namespace blast
#endif // BLDRIVER_H
